--[[----------------------------------------------------------------------------
-- Duplex.Mixer
----------------------------------------------------------------------------]]--

--[[

  About

  Mixer is a generic class for controlling the Renoise mixer
  - supported on a wide variety of hardware
  - supports faders, buttons or dials for input

  
  ---------------------------------------------------------

  Control-map assignments 

  "master"  - specify this to control the master track seperately
  "page" - specify this to control the track offset
  "mode" - specify this to control if pre of post fx values are mapped

  [][][][] []  [][][]     |
  |||||||| []  [][][]     |
  |||||||| []  [][][]   "levels" (optional)
  |||||||| []  [][][]     |
  [][][][] []  [][][]     |
                   
  [][][][] []  [][][]   "mute" (optional)
                   
  [][][][] []  [][][]   "panning" (optional)

  normal  mst   send
  tracks track tracks


  ---------------------------------------------------------
  
  Navigation features

  The mixer will automatically follow the global position
  if the option is enabled. It is displayed like this:

  Four-track mixer, where we select track 2/8:
  The mixer is divided into two "pages"

     Page#1    |    Page#2
  [1][2][3][4] | [5][6][7][8] <-- tracks
      x        |           x  <-- track offset (2/8)

  Eight-track mixer, where we select track 2/8:
  there is enough room for all tracks at once

     Page#1
  [1][2][3][4][5][6][7][8]    <-- tracks
      x                 x     <-- track offset (2/8)


  ---------------------------------------------------------
  
  Pre/Post FX Volume/Pan
  
  The mixer allows controlling both, pre and post FX values. 
  By default its configured to use post FX values. 
  
  You can either change the default via the controller options,
  or by mapping something to the "mode" group, in order to 
  change the mode dynamically from the controller
  
--]]


--==============================================================================

class 'Mixer' (Application)

function Mixer:__init(display,mappings,options)
  TRACE("Mixer:__init",display,mappings,options)

  -- constructor 
  Application.__init(self)

  self.display = display

    -- define the options (with defaults)

  self.ALL_TRACKS = 1
  self.NORMAL = 2
  self.NORMAL_MASTER = 3
  self.MASTER = 4
  self.MASTER_SEND = 5
  self.SEND = 6

  self.MODE_PREFX = 1
  self.MODE_POSTFX = 2

  self.MUTE_NORMAL = 1
  self.MUTE_INVERTED = 2

  self.options = {
    pre_post = {
      label = "Mode",
      description = "Change if either Pre or Post FX volume/pan is controlled",
      items = {
        "Pre FX volume and panning",
        "Post FX volume and panning",
      },
      default = 1
    },
    invert_mute = {
      label = "Invert mute",
      description = "Decide how to display muted tracks",
      items = {
        "Button is lit when track is muted",
        "Button is lit when track is active",
      },
      default = 2
    },

    --[[
    -- TODO
    include_tracks = {
      label = "Tracks",
      description = "Select any combination of tracks that you want to " ..
        "include: normal, master and send tracks.",
      items = {
        "Include all tracks",
        "Normal tracks only",
        "Normal + master tracks",
        "Master track only",
        "Master + send tracks",
        "Send tracks only",
      },
      default = 1,
    },
    track_offset = {
      label = "Offset",
      description = "Change the offset if you want this Mixer to begin " .. 
        "with a different track",
      items = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16},
      default = 1,
    },
    sync_position = {
      label = "Sync to global position",
      description = "Set to true if you want the Mixer to align with the " ..
        "selected track in Renoise",
      items = {true,false},
      default = 1,
    },
    ]]
  }

  -- define control-maps groups 
  self.mappings = {
    master = {
      description = "Master volume",
      ui_component = UI_COMPONENT_SLIDER,
    },
    levels = {
      description = "Track volume",
      ui_component = UI_COMPONENT_SLIDER,
      greedy = true,
    },
    panning = {
      description = "Track panning",
      ui_component = UI_COMPONENT_SLIDER,
      greedy = true,
    },
    mute = {
      description = "Mute track",
      ui_component = UI_COMPONENT_TOGGLEBUTTON,
      greedy = true,
    },
    solo = {
      description = "Solo track",
      ui_component = UI_COMPONENT_TOGGLEBUTTON,
      greedy = true,
    },
    page = {
      description = "Mixer track navigator",
      ui_component = UI_COMPONENT_SPINNER,
    },
    mode = {
      description = "Mixer Pre/Post FX mode",
      ui_component = UI_COMPONENT_TOGGLEBUTTON,
    },
  }

  -- default palette: should degrade nicely for all supported devices

  self.palette = {
    background = {
      color={0,0,0},
      text="·",
    },
      -- normal tracks are green
    normal_tip = {
      color={0x00,0xff,0xff},
      text="■",
    },
    normal_tip_dimmed = {
      color={0x00,0x40,0xff},
      text="□",
    },
    normal_lane = {
      color={0x00,0x81,0xff},
      text="▪",
    },
    normal_lane_dimmed = {
      color={0x00,0x40,0xff},
      text="▫",
    },
    normal_mute = {
      color={0x00,0xff,0xff},
      text="■",
    },
      -- master track is yellow
    master_tip = {
      color={0xff,0xff,0xff},
      text="■",
    },
    master_lane = {
      color={0x80,0x80,0xff},
      text="▪",
    },
    -- send tracks are red
    send_tip = {
      color={0xff,0x00,0xff},
      text="■",
    },
    send_tip_dimmed = {
      color={0x40,0x00,0xff},
      text="□",
    },
    send_lane = {
      color={0x81,0x00,0xff},
      text="▪",
    },
    send_lane_dimmed = {
      color={0x40,0x00,0xff},
      text="▫",
    },
    send_mute = {
      color={0xff,0x00,0xff},
      text="■",
    },
  }

  -- the various controls
  self.__master = nil
  self.__levels = nil
  self.__panning = nil
  self.__mutes = nil
  self.__solos = nil
  self.__page_control = nil
  self.__mode_control = nil

  self.__width = 4
  self.__height = 1

  -- todo: extract this from options
  self.__include_normal = true
  self.__include_master = true
  self.__include_send = true

  -- offset of the whole track mapping, controlled by the track navigator
  self.__track_offset = 0
  
  -- current track properties we are listening to
  self.__attached_track_observables = table.create()

  -- apply arguments
  self:__apply_options(options)
  self:__apply_mappings(mappings)

  -- toggle, which defines if we're controlling the pre or post fx vol/pans
  self.__postfx_mode = (self.options.pre_post.value == self.MODE_POSTFX)

end


--------------------------------------------------------------------------------

-- volume level changed from Renoise

function Mixer:set_track_volume(control_index, value)
  TRACE("Mixer:set_track_volume", control_index, value)

  if (self.active) then
    if (self.__levels ~= nil) then
      self.__levels[control_index]:set_value(value)
      
      -- update the master as well, if it has its own UI representation
      if (self.__master ~= nil) and 
         (control_index + self.__track_offset == get_master_track_index()) 
      then
        self.__master:set_value(value)
      end
    end
  end
end


--------------------------------------------------------------------------------

-- panning changed from Renoise

function Mixer:set_track_panning(control_index, value)
  TRACE("Mixer:set_track_panning", control_index, value)

  if (self.active and self.__panning ~= nil) then
    self.__panning[control_index]:set_value(value)
  end
end


--------------------------------------------------------------------------------

-- mute state changed from Renoise

function Mixer:set_track_mute(control_index, state)
  TRACE("Mixer:set_track_mute", control_index, state)

  if (self.active and self.__mutes ~= nil) then
    -- set mute state to the button
    local active = (state == MUTE_STATE_ACTIVE)
    self.__mutes[control_index]:set(active)

    -- make controls appear dimmed, to show that the track is inactive
    if (self.__levels) then
      self.__levels[control_index]:set_dimmed(not active)
    end
    
    if (self.__panning) then
      self.__panning[control_index]:set_dimmed(not active)
    end
  end
end


--------------------------------------------------------------------------------

-- solo state changed from Renoise

function Mixer:set_track_solo(control_index, state)
  TRACE("Mixer:set_track_solo", control_index, state)

  if (self.active and self.__solos ~= nil) then
    -- set mute state to the button
    self.__solos[control_index]:set(state)
  end
end


--------------------------------------------------------------------------------

-- update: set all controls to current values from renoise

function Mixer:update()  
  TRACE("Mixer:update()")

  local master_track_index = get_master_track_index()
  local tracks = renoise.song().tracks

  -- track volume/panning/mute and solo
  
  for control_index = 1,self.__width do
  
    local track_index = self.__track_offset + control_index
    local track = tracks[track_index]
      
    -- set default values
    if (track_index <= #tracks) then
      
      -- update component states from the track
      if (self.__postfx_mode) then
        self:set_track_volume(control_index, track.postfx_volume.value)
        self:set_track_panning(control_index, track.postfx_panning.value)
      else
        self:set_track_volume(control_index, track.prefx_volume.value)
        self:set_track_panning(control_index, track.prefx_panning.value)
      end
      
      -- show that we can't change the master mute state
      if (track_index == master_track_index) then
        self:set_track_mute(control_index, MUTE_STATE_ACTIVE)
      else
        self:set_track_mute(control_index, track.mute_state)
      end
      
       self:set_track_solo(control_index, track.solo_state)
    else
      -- deactivate, reset controls which have no track
      self:set_track_volume(control_index, 0)
      self:set_track_panning(control_index, 0)
      self:set_track_mute(control_index, MUTE_STATE_OFF)
      self:set_track_solo(control_index, false)
    end

    -- apply palette to controls

    local track_palette = {}
    local mute_palette = {}

    if (track_index < master_track_index) then
      -- normal tracks
      track_palette.tip           = self.palette.normal_tip
      track_palette.tip_dimmed    = self.palette.normal_tip_dimmed
      track_palette.track         = self.palette.normal_lane
      track_palette.track_dimmed  = self.palette.normal_lane_dimmed
      mute_palette.foreground     = self.palette.normal_mute
    elseif (track_index == master_track_index) then
      -- master track
      track_palette.tip           = self.palette.master_tip
      track_palette.track         = self.palette.master_lane
      mute_palette.foreground     = self.palette.background
    elseif (track_index <= #tracks) then
      -- send tracks
      track_palette.tip           = self.palette.send_tip
      track_palette.tip_dimmed    = self.palette.send_tip_dimmed
      track_palette.track         = self.palette.send_lane
      track_palette.track_dimmed  = self.palette.send_lane_dimmed
      mute_palette.foreground     = self.palette.send_mute

    else 
      -- unmapped tracks are black

    end

    if (self.__levels) then
      self.__levels[control_index]:set_palette(track_palette)
    end
    
    if (self.__panning) then
      self.__panning[control_index]:set_palette(track_palette)
    end

    if (self.__mutes) then
      self.__mutes[control_index]:set_palette(mute_palette)
    end

    if (self.__solos) then
      self.__solos[control_index]:set_palette(mute_palette)
    end
  end
  
  
  -- master volume
  
  if (self.__master ~= nil) then
     if (self.__postfx_mode) then
       self.__master:set_value(get_master_track().postfx_volume.value)
     else
       self.__master:set_value(get_master_track().prefx_volume.value)
     end
  end
  
  
  -- page controls

  if (self.__page_control) then
    self.__page_control:set_index(self.__track_offset)
  end


  -- mode controls

  if (self.__mode_control) then
    self.__mode_control:set(self.__postfx_mode)
  end

end


--------------------------------------------------------------------------------

-- start/resume application

function Mixer:start_app()
  TRACE("Mixer.start_app()")

  if not (self.__created) then 
    self:__build_app()
  end

  Application.start_app(self)
  self:update()
end


--------------------------------------------------------------------------------

function Mixer:on_new_document()
  TRACE("Mixer:on_new_document")
  
  self:__attach_to_song()
  
  if (self.active) then
    self:update()
  end
end


--------------------------------------------------------------------------------

-- build_app: create a grid or fader/encoder layout

function Mixer:__build_app()
  TRACE("Mixer:__build_app(")

  Application.__build_app(self)

  -- check if the control-map describes a grid controller
  -- (slider is composed from individual buttons in grid mode)
  
  local slider_grid_mode = false
  
  local control_map_groups = self.display.device.control_map.groups 
  local levels_group = control_map_groups[self.mappings.levels.group_name]

  if (levels_group) then
    for attr, param in pairs(levels_group) do
      if (attr == "xarg" and param["columns"]) then
        slider_grid_mode = true
        self.__width = tonumber(param["columns"])
        self.__height = math.ceil(#levels_group / self.__width)
        break
      end
    end
  end
  
  local embed_mutes = (self.mappings.mute.group_name == 
    self.mappings.levels.group_name)
  
  local embed_master = (self.mappings.master.group_name == 
    self.mappings.levels.group_name)

  if (slider_grid_mode) then
    if (embed_master) then
      self.__width = self.__width-1
    end

  else
    -- extend width to the number of parameters in the levels group
    if (levels_group) then
      self.__width = #levels_group
    end
  end

  
  -- construct the display
  
  self.__levels = (self.mappings.levels.group_name) and {} or nil
  self.__panning = (self.mappings.panning.group_name) and {} or nil
  self.__mutes = (self.mappings.mute.group_name) and {} or nil
  self.__solos = (self.mappings.solo.group_name) and {} or nil
  
  self.__master = nil
  self.__page_control = nil
  self.__mode_control = nil
  
  for control_index = 1,self.__width do

    -- sliders --------------------------------------------

    if (self.mappings.levels.group_name) then

      local y_pos = (embed_mutes) and 2 or 1
      local c = UISlider(self.display)
      c.group_name = self.mappings.levels.group_name
      c.tooltip = self.mappings.levels.description
      c:set_pos(control_index,y_pos)
      c.toggleable = true
      c.flipped = false
      c.ceiling = RENOISE_DECIBEL
      c.orientation = VERTICAL
      c:set_size(self.__height-(y_pos-1))
  
      -- slider changed from controller
      c.on_change = function(obj) 

        local track_index = self.__track_offset + control_index
  
        if (not self.active) then
          return false
          
        elseif (track_index == get_master_track_index()) then
          if (self.__master) then
            -- this will cause another event...
            self.__master:set_value(obj.value)
          else
            local track = renoise.song().tracks[track_index]
            
            local volume = (self.__postfx_mode) and 
              track.postfx_volume or track.prefx_volume
        
            volume.value = obj.value
          end
          return true
        
        elseif (track_index > #renoise.song().tracks) then
          -- track is outside bounds
          return false

        else
          local track = renoise.song().tracks[track_index]
          
          local volume = (self.__postfx_mode) and 
            track.postfx_volume or track.prefx_volume
        
          volume.value = obj.value
        end
      end
      
      self:__add_component(c)
      self.__levels[control_index] = c

    end
    

    -- encoders -------------------------------------------

    if (self.mappings.panning.group_name) then
      local c = UISlider(self.display)
      c.group_name = self.mappings.panning.group_name
      c.tooltip = self.mappings.panning.description
      c:set_pos(control_index)
      c.toggleable = true
      c.flipped = false
      c.ceiling = 1.0
      c.orientation = VERTICAL
      c:set_size(1)
      
      -- slider changed from controller
      c.on_change = function(obj) 
        local track_index = self.__track_offset + control_index
  
        if (not self.active) then
          return false
  
        elseif (track_index > #renoise.song().tracks) then
          -- track is outside bounds
          return false
  
        else
          local track = renoise.song().tracks[track_index]
          
          local panning = (self.__postfx_mode) and 
            track.postfx_panning or track.prefx_panning
        
          panning.value = obj.value
          
          return true
        end
      end
      
      self:__add_component(c)
      self.__panning[control_index] = c

    end
        
     
    -- mute buttons -----------------------------------
    
    if (self.mappings.mute.group_name) then
      local c = UIToggleButton(self.display)
      c.group_name = self.mappings.mute.group_name
      c.tooltip = self.mappings.mute.description
      c:set_pos(control_index)
      c.inverted = (self.options.invert_mute.value == self.MUTE_NORMAL) or false
      c.active = false
  
      -- mute state changed from controller
      -- (update the slider.dimmed property)
      c.on_change = function(obj) 
        local track_index = self.__track_offset + control_index
        
        if (not self.active) then
          return false
        
        elseif (track_index == get_master_track_index()) then
          -- can't mute the master track
          return false
        
        elseif (track_index > #renoise.song().tracks) then
          -- track is outside bound
          return false
        end
        
        local track = renoise.song().tracks[track_index]

        if (obj.active) then
          track:unmute()
        else 
          track:mute()
        end

        if (self.__levels) then
          self.__levels[control_index]:set_dimmed(not obj.active)
        end
        
        if (self.__panning) then
          self.__panning[control_index]:set_dimmed(not obj.active)
        end
        
        return true
      end
      
      self:__add_component(c)
      self.__mutes[control_index] = c    

    end
    

    -- solo buttons -----------------------------------

    if (self.mappings.solo.group_name) then
      local c = UIToggleButton(self.display)
      c.group_name = self.mappings.solo.group_name
      c.tooltip = self.mappings.solo.description
      c:set_pos(control_index)
      c.inverted = false
      c.active = false
  
      -- mute state changed from controller
      -- (update the slider.dimmed property)
      c.on_change = function(obj) 
        local track_index = self.__track_offset + control_index
        
        if (not self.active) then
          return false
        
        elseif (track_index > #renoise.song().tracks) then
          -- track is outside bound
          return false
        end
        
        local track = renoise.song().tracks[track_index]
        track.solo_state = obj.active
  
        return true
      end
      
      self:__add_component(c)
      self.__solos[control_index] = c    

    end
  end
  

  -- master fader ------------------------------

  if (self.mappings.master.group_name) then
    
    local c = UISlider(self.display)
    c.group_name = self.mappings.master.group_name
    c.tooltip = self.mappings.master.description
    c:set_pos((embed_master) and (self.__width + 1) or 1)
    c.toggleable = true
    c.ceiling = RENOISE_DECIBEL
    c:set_size(self.__height)
    
    c.on_change = function(obj) 
      if (not self.active) then
        return false
      else
        local master_control_index = 
          get_master_track_index() - self.__track_offset
        
        if (self.__levels and 
            master_control_index > 0 and 
            master_control_index <= self.__width) 
        then
          -- this will cause another event...
          self.__levels[master_control_index]:set_value(obj.value)

        else
          local volume = (self.__postfx_mode) and 
            get_master_track().postfx_volume or 
            get_master_track().prefx_volume
        
          volume.value = obj.value
        end
        
        return true
      end
    end 
    
    self:__add_component(c)
    self.__master = c

  end
  
  
  -- track scrolling ---------------------------

  if (self.mappings.page.group_name) then
  
    local c = UISpinner(self.display)
    c.group_name = self.mappings.page.group_name
    c.tooltip = self.mappings.page.description
    c.index = 0
    c.step_size = self.__width
    c.minimum = 0
    c.maximum = math.max(0, 
      #renoise.song().tracks - self.__width)
    c:set_pos(self.mappings.page.index or 1)
    c.text_orientation = HORIZONTAL

    c.on_change = function(obj) 
      if (not self.active) then
        return false
      end

      self.__track_offset = obj.index
      
      local new_song = false
      self:__attach_to_tracks(new_song)
      
      self:update()

      return true
    end
    
    self:__add_component(c)
    self.__page_control = c

  end


  -- Pre/Post FX mode ---------------------------

  if (self.mappings.mode.group_name) then
    local c = UIToggleButton(self.display)
    c.group_name = self.mappings.mode.group_name
    c.tooltip = self.mappings.mode.description
    c:set_pos(self.mappings.mode.index or 1)
    c.inverted = false
    c.active = false

    -- mode state changed from controller
    c.on_change = function(obj) 
      if (not self.active) then
        return false
      end
      
      if (self.__postfx_mode ~= obj.active) then
        self.__postfx_mode = obj.active
        
        local new_song = false
        self:__attach_to_tracks(new_song)
        
        self:update()
      end

      return true
    end
    
    self:__add_component(c)
    self.__mode_control = c

  end
    
  -- the finishing touches
  self:__attach_to_song()
end


--------------------------------------------------------------------------------

-- adds notifiers to song
-- invoked when a new document becomes available

function Mixer:__attach_to_song()
  TRACE("Mixer:__attach_to_song")
  
  -- update on track changes in the song
  renoise.song().tracks_observable:add_notifier(
    function()
      TRACE("Mixer:tracks_changed fired...")
      
      local new_song = false
      self:__attach_to_tracks(new_song)
      
      if (self.active) then
        self:update()
      end
    end
  )

  -- and immediately attach to the current track set
  local new_song = true
  self:__attach_to_tracks(new_song)
end


--------------------------------------------------------------------------------

-- add notifiers to parameters
-- invoked when tracks are added/removed/swapped

function Mixer:__attach_to_tracks(new_song)
  TRACE("Mixer:__attach_to_tracks", new_song)

  local tracks = renoise.song().tracks

  -- validate and update the sequence/track offset
  if (self.__page_control) then
    self.__page_control:set_range(nil,math.max(0, 
      #renoise.song().tracks - self.__width))
  end
    
  -- detach all previously added notifiers first
  -- but don't even try to detach when a new song arrived. old observables
  -- will no longer be alive then...
  if (not new_song) then
    for _,observable in pairs(self.__attached_track_observables) do
      -- temp security hack. can also happen when removing tracks
      pcall(function() observable:remove_notifier(self) end)
    end
  end
  
  self.__attached_track_observables:clear()
  
  
  -- attach to the new ones in the order we want them
  local master_done = false
  local master_idx = get_master_track_index()
  
  for control_index = 1,math.min(#tracks, self.__width) do
    local track_index = self.__track_offset + control_index
    local track = tracks[track_index]

    if (track_index == master_idx) then
      master_done = true
    end
    
    -- track volume level 
    if (self.__levels) then
      local volume = (self.__postfx_mode) and 
        track.postfx_volume or track.prefx_volume
      
      self.__attached_track_observables:insert(
        volume.value_observable)
      
      volume.value_observable:add_notifier(
        self, 
        function()
          if (self.active) then
            local value = volume.value
            -- compensate for potential loss of precision 
            if not compare(self.__levels[control_index].value, value, 1000) then
              self:set_track_volume(control_index, value)
            end
          end
        end 
      )
    end
    
    -- track panning level 
    if (self.__panning) then
      local panning = (self.__postfx_mode) and 
         track.postfx_panning or track.prefx_panning
      
      self.__attached_track_observables:insert(
        panning.value_observable)

      panning.value_observable:add_notifier(
        self, 
        function()
          if (self.active) then
            local value = panning.value
            -- compensate for potential loss of precision 
            if not compare(self.__panning[control_index].value, value, 1000) then
              self:set_track_panning(control_index, value)
            end
          end
        end 
      )
    end
    
    -- track mute-state 
    if (self.__mutes) then
      self.__attached_track_observables:insert(track.mute_state_observable)

      track.mute_state_observable:add_notifier(
        self, 
        function()
          if (self.active) then
            self:set_track_mute(control_index, track.mute_state)
          end
        end 
      )
    end
    
    -- track solo-state 
    if (self.__solos) then
      self.__attached_track_observables:insert(track.solo_state_observable)

      track.solo_state_observable:add_notifier(
        self, 
        function()
          if (self.active) then
            self:set_track_solo(control_index, track.solo_state)
          end
        end 
      )
    end
  end

  -- if master wasn't already mapped just before
  if (not master_done and self.__master) and 
    (self.__include_master) then
    local track = renoise.song().tracks[master_idx]
    
    local volume = (self.__postfx_mode) and 
      track.postfx_volume or track.prefx_volume

    self.__attached_track_observables:insert(
      volume.value_observable)
  
    volume.value_observable:add_notifier(
      self, 
      function()
        if (self.active) then
          local value = volume.value
          -- compensate for potential loss of precision 
          if not compare(self.__master.value, value, 1000) then
            self.__master:set_value(value)
          end
        end
      end 
    )
  end

end

